#include "fosfft.h"

#include <stdlib.h>
#include <stdint.h>
#include <math.h>

/*
 *  struct fosfft_struct for type `fosfft_t` as declared in `fosfft.h`
 */
struct fosfft_struct {
  int max_n;                 // maximum `n` value supported
  complex double factors[];  // twiddle factors
};
struct fosfftf_struct {
  int max_n;
  complex float factors[];
};

/*
 *  implementation of fosfft_init() as declared in `fosfft.h`
 */
fosfft_t *fosfft_init(int max_n, bool reverse) {
  fosfft_t *env;
  int i;
  // check integer overflow for memory size calculation:
  if (max_n/2 > (SIZE_MAX - sizeof(fosfft_t))/sizeof(*env->factors))
    return NULL;
  // allocate memory:
  env = calloc(1, sizeof(*env) + (max_n/2)*sizeof(*env->factors));
  if (!env) return NULL;
  // fill and return struct:
  env->max_n = max_n;
  if (reverse) {
    for (i=0; i<max_n/2; i++) env->factors[i] = cexp(2*M_PI/max_n*i*I);
  } else {
    for (i=0; i<max_n/2; i++) env->factors[i] = cexp(-2*M_PI/max_n*i*I);
  }
  return env;
}
fosfftf_t *fosfftf_init(int max_n, bool reverse) {
  fosfftf_t *env;
  int i;
  // check integer overflow for memory size calculation:
  if (max_n/2 > (SIZE_MAX - sizeof(fosfftf_t))/sizeof(*env->factors))
    return NULL;
  // allocate memory:
  env = calloc(1, sizeof(*env) + (max_n/2)*sizeof(*env->factors));
  if (!env) return NULL;
  // fill and return struct:
  env->max_n = max_n;
  if (reverse) {
    for (i=0; i<max_n/2; i++) env->factors[i] = cexp(2*M_PI/max_n*i*I);
  } else {
    for (i=0; i<max_n/2; i++) env->factors[i] = cexp(-2*M_PI/max_n*i*I);
  }
  return env;
}

/*
 *  macro for function headers of several recursively called functions
 */
#define fosfft_func(name) \
  static inline void name( \
  const complex double *restrict src, int srcstep, \
  complex double *restrict dst, int n, \
  const complex double *restrict factors, int factorstep )
#define fosfftf_func(name) \
  static inline void name( \
  const complex float *restrict src, int srcstep, \
  complex float *restrict dst, int n, \
  const complex float *restrict factors, int factorstep )

/*
 *  combine two processed sub-blocks into one
 */
fosfft_func(fosfft_postproc) {
  int i;
  complex double u, v;
  for (i=0; i<n/2; i++) {
    u = dst[i];
    v = dst[n/2+i] * factors[factorstep*i];
    dst[i] = u + v;
    dst[n/2+i] = u - v;
  }
}
fosfftf_func(fosfftf_postproc) {
  int i;
  complex float u, v;
  for (i=0; i<n/2; i++) {
    u = dst[i];
    v = dst[n/2+i] * factors[factorstep*i];
    dst[i] = u + v;
    dst[n/2+i] = u - v;
  }
}

/*
 *  forward declaration for cyclic dependent functions
 */
fosfft_func(fosfft_lvl1);
fosfftf_func(fosfftf_lvl1);

/*
 *  calculate FFT or IFFT for certain `n` where recursion might stop
 */
fosfft_func(fosfft_lvl2) {
  if (n == 4) {
    // stop recursion and calculate directly:
    dst[0] = src[0] + src[srcstep*2];
    dst[1] = src[0] - src[srcstep*2];
    dst[2] = src[srcstep] + src[srcstep*3];
    dst[3] = src[srcstep] - src[srcstep*3];
  } else {
    // use recursion for calculation:
    fosfft_lvl1(src, srcstep*2, dst, n/2, factors, factorstep*2);
    fosfft_lvl1(src+srcstep, srcstep*2, dst+n/2, n/2, factors, factorstep*2);
  }
  // combine results into one:
  fosfft_postproc(src, srcstep, dst, n, factors, factorstep);
}
fosfftf_func(fosfftf_lvl2) {
  if (n == 4) {
    // stop recursion and calculate directly:
    dst[0] = src[0] + src[srcstep*2];
    dst[1] = src[0] - src[srcstep*2];
    dst[2] = src[srcstep] + src[srcstep*3];
    dst[3] = src[srcstep] - src[srcstep*3];
  } else {
    // use recursion for calculation:
    fosfftf_lvl1(src, srcstep*2, dst, n/2, factors, factorstep*2);
    fosfftf_lvl1(src+srcstep, srcstep*2, dst+n/2, n/2, factors, factorstep*2);
  }
  // combine results into one:
  fosfftf_postproc(src, srcstep, dst, n, factors, factorstep);
}

/*
 *  calculation of FFT or IFFT for other values of `n` where recursion can not
 *  stop yet; using different functions allows the compiler to inline all but
 *  one of the fosfft_lvl* functions
 */
#define fosfft_recursive(name, subcall) \
  fosfft_func(name) { \
    subcall(src, srcstep*2, dst, n/2, factors, factorstep*2); \
    subcall(src+srcstep, srcstep*2, dst+n/2, n/2, factors, factorstep*2); \
   fosfft_postproc(src, srcstep, dst, n, factors, factorstep); \
 }
#define fosfftf_recursive(name, subcall) \
  fosfftf_func(name) { \
    subcall(src, srcstep*2, dst, n/2, factors, factorstep*2); \
    subcall(src+srcstep, srcstep*2, dst+n/2, n/2, factors, factorstep*2); \
   fosfftf_postproc(src, srcstep, dst, n, factors, factorstep); \
 }
fosfft_recursive(fosfft_lvl3, fosfft_lvl2)
fosfft_recursive(fosfft_lvl0, fosfft_lvl3)
fosfft_recursive(fosfft_lvl1, fosfft_lvl0)
fosfftf_recursive(fosfftf_lvl3, fosfftf_lvl2)
fosfftf_recursive(fosfftf_lvl0, fosfftf_lvl3)
fosfftf_recursive(fosfftf_lvl1, fosfftf_lvl0)

/*
 *  macro used by fosfft() function for calling different functions to begin
 *  recursion
 */
#define fosfft_call_level(name) \
  name(src, 1, dst, n, env->factors, env->max_n / n); break

/*
 *  implementation of fosfft() as declared in `fosfft.h`
 */
void fosfft(
  fosfft_t *env,
  const complex double *restrict src,
  complex double *restrict dst,
  int n
) {
  // handle cases n==1 and n==2:
  if (n == 1) dst[0] = src[0];
  else if (n == 2) { dst[0] = src[0] + src[1]; dst[1] = src[0] - src[1]; }
  // handle case of n >= 4:
  else {
    // determine which recursive function to start with:
    int m = n;
    while (!(m & ((1<<4)-1))) m >>= 4;
    // start recursion with appropriate function:
    switch (m) {
      case 1<<0: fosfft_call_level(fosfft_lvl0);
      case 1<<1: fosfft_call_level(fosfft_lvl1);
      case 1<<2: fosfft_call_level(fosfft_lvl2);
      case 1<<3: fosfft_call_level(fosfft_lvl3);
    }
  }
}
void fosfftf(
  fosfftf_t *env,
  const complex float *restrict src,
  complex float *restrict dst,
  int n
) {
  // handle cases n==1 and n==2:
  if (n == 1) dst[0] = src[0];
  else if (n == 2) { dst[0] = src[0] + src[1]; dst[1] = src[0] - src[1]; }
  // handle case of n >= 4:
  else {
    // determine which recursive function to start with:
    int m = n;
    while (!(m & ((1<<4)-1))) m >>= 4;
    // start recursion with appropriate function:
    switch (m) {
      case 1<<0: fosfft_call_level(fosfftf_lvl0);
      case 1<<1: fosfft_call_level(fosfftf_lvl1);
      case 1<<2: fosfft_call_level(fosfftf_lvl2);
      case 1<<3: fosfft_call_level(fosfftf_lvl3);
    }
  }
}

/*
 *  implementation of fosfft_kaiser() as declared in `fosfft.h`;
 *  for -1 <= x <= 1, the return value equals to
 *  Bessel_I0(beta * sqrt(1.0 - x*x)) where Bessel_I0 is the
 *  modified Bessel function of first kind of order zero;
 *  for abs(x) > 1.0, the result is undefined
 */
double fosfft_kaiser(double beta, double x) {
  int j; double a, f = 1.0, d = 1.0, y = 1.0;
  // calculate argument for Bessel function:
  x = beta * sqrt(1.0 - x*x);
  // calculate modified Bessel function of first kind of order zero:
  if (isnan(x)) return NAN;
  x = x*x/4.0;
  for (j=1; ; j++) {
    f *= x; if (!isfinite(f)) return INFINITY; d *= j*j;
    a = f/d; if (!a) break; y += a;
  }
  return y;
}

