#pragma once

#include <stdbool.h>
#include <complex.h>

/*  All types and functions with `fosfftf` prefix operate on floats instead of
 *  doubles.
 */

/*
 *  opaque data structure holding pre-calculated values for executing FFTs
 *  later
 */
typedef struct fosfft_struct fosfft_t;
typedef struct fosfftf_struct fosfftf_t;

/*  pre-calculate values for executing FFTs (or IFFTs) later; returned
 *  structure shall be free()ed by caller when not needed anymore;
 *  `max_n` _MUST_ be a power of 2 (elsewise behavior is undefined!)
 */
fosfft_t *fosfft_init(int max_n, bool reverse);
fosfftf_t *fosfftf_init(int max_n, bool reverse);
#define fosfft_init_forward(max_n) fosfft_init((max_n), false)
#define fosfft_init_reverse(max_n) fosfft_init((max_n), true)
#define fosfftf_init_forward(max_n) fosfftf_init((max_n), false)
#define fosfftf_init_reverse(max_n) fosfftf_init((max_n), true)

/*  perform FFT (or IFFT, depending on initialization of `env`);
 *  `src` and `dst` must point to non-overlapping memory;
 *  factor 1/n is included neither in FFT nor in IFFT and must be
 *  applied by the caller where necessary;
 *  `n` _MUST_ be a power of 2 and equal to or less than `max_n` as
 *  passed to initialization function (elsewise behavior is undefined!)
 */
void fosfft(
  fosfft_t *env,
  const complex double *restrict src,
  complex double *restrict dst,
  int n
);
void fosfftf(
  fosfftf_t *env,
  const complex float *restrict src,
  complex float *restrict dst,
  int n
);

/*
 *  return value of Kaiser window with parameter `beta` at position
 *  -1.0 <= x <= 1.0; result is not normalized, i.e. multiplied with an
 *  unknown constant factor; divide by fosfft_kaiser(0) to normalize to
 *  values equal to or smaller than 1.0; result is undefined for abs(x) > 1.0
 */
double fosfft_kaiser(double beta, double x);

